ALL_FEATURES = {
    "create_with_stat": (3, 5, 0),
    "containers": (3, 5, 1),
    "reconfigure": (3, 5, 0),
}


class Features(object):

    def __init__(self, version_info):
        for feature_name, version_introduced in ALL_FEATURES.items():
            setattr(self, feature_name, version_info >= version_introduced)



from aiozk import WatchEvent
from aiozk.exc import NoNode

from .base_watcher import BaseWatcher


class DataWatcher(BaseWatcher):

    watched_events = [WatchEvent.DATA_CHANGED, WatchEvent.DELETED]

    def __init__(self, *args, wait_for_create=False, **kwargs):
        super().__init__(*args, **kwargs)

        if wait_for_create:
            self.watched_events.append(WatchEvent.CREATED)

    async def fetch(self, path):
        # exists() gives create, delete, and update watches
        watch_via_exists = WatchEvent.CREATED in self.watched_events
        if watch_via_exists:
            exists = await self.client.exists(path, watch=True)
            if not exists:
                raise NoNode
        data = await self.client.get_data(path=path, watch=not watch_via_exists)
        return data



from aiozk import exc

from .base_lock import BaseLock


class SharedLock(BaseLock):

    async def acquire_read(self, timeout=None):
        result = None
        while not result:
            try:
                result = await self.wait_in_line(
                    "read", timeout, blocked_by=("write")
                )
            except exc.SessionLost:
                continue
        return result

    async def acquire_write(self, timeout=None):
        result = None
        while not result:
            try:
                result = await self.wait_in_line("write", timeout)
            except exc.SessionLost:
                continue
        return result



import asyncio
import logging
import uuid
import pytest

from aiozk import WatchEvent
from .conftest import get_client


logging.getLogger('asyncio').setLevel(logging.DEBUG)


@pytest.fixture
async def full_zk(zk, path):
    child_1 = f'{path}/{uuid.uuid4().hex}'
    child_2 = f'{path}/{uuid.uuid4().hex}'
    await zk.create(path)
    await zk.create(child_1)
    await zk.create(child_2)
    yield zk
    await zk.delete(child_2)
    await zk.delete(child_1)
    await zk.delete(path)


@pytest.mark.asyncio
async def test_children(full_zk, path):
    resp = await full_zk.get_children(path)
    assert len(resp) == 2


@pytest.mark.asyncio
async def test_cancel_crash(zk, path):
    async def wait_loop():
        while 1:
            await zk.wait_for_events([WatchEvent.DATA_CHANGED], path)

    f = asyncio.ensure_future(wait_loop())
    f.cancel()


@pytest.mark.asyncio
async def test_closed_close():
    zk = get_client()
    await asyncio.wait_for(zk.session.close(), 2)



@pytest.mark.asyncio
async def test_inconsistent_zxid():
    async def coro():
        zk = get_client()
        await zk.start()
        # simulate failed connection
        await zk.session.close()
        zk.session.last_zxid = 1231231241312312
        await zk.session.start()
        await zk.session.close()
    try:
        await asyncio.wait_for(coro(), timeout=10)
    except asyncio.TimeoutError as exc:
        pytest.fail("Failed with timeout on session reconnection attemt")


@pytest.mark.asyncio
async def test_session_reconnect():
    async def coro():
        zk = get_client()
        await zk.start()
        await zk.session.close()
        await zk.session.start()
        await zk.session.close()
    try:
        await asyncio.wait_for(coro(), timeout=10)
    except asyncio.TimeoutError as exc:
        pytest.fail("Failed with timeout on session reconnection attemt")


@pytest.mark.asyncio
async def test_raw_get(full_zk, path):
    # type: (aiozk.ZKClient, str) -> None
    """Test that get returns data + stat"""
    data, stat = await full_zk.get(path)
    assert data is None
    assert stat.version == 0

@pytest.mark.asyncio
async def test_raw_set(full_zk, path):
    # type: (aiozk.ZKClent, str) -> None
    """Test that raw set returns Stat"""
    stat = await full_zk.set(path, 'asdf', -1)
    assert stat.data_length == 4
    assert stat.version == 1


from .request import Request
from .response import Response
from .primitives import UString


class SyncRequest(Request):
    """
    """
    opcode = 9

    parts = (
        ("path", UString),
    )


class SyncResponse(Response):
    """
    """
    opcode = 9

    parts = (
        ("path", UString),
    )



import asyncio
import logging

import pytest

from ..recipes.counter import Counter  # noqa


async def _simple_helper(zk, path, val):
    counter = zk.recipes.Counter(path, default=val)
    await counter.start()
    _type = type(val)
    assert counter.value == val
    assert type(counter.value) == _type
    for _i in range(4):
        await counter.incr()
    await zk.delete(path)
    counter.stop()
    assert counter.value == 4


@pytest.mark.asyncio
async def test_counter_simple_int(zk, path):
    await _simple_helper(zk, path, 0)


@pytest.mark.asyncio
async def test_counter_simple_float(zk, path):
    await _simple_helper(zk, path, 0.0)


@pytest.mark.asyncio
async def test_counter_multiple(zk, path):
    async def worker():
        c = zk.recipes.Counter(path)
        await c.start()
        await c.incr()

    workers = []
    for _i in range(5):
        workers.append(worker())

    done, _pending = await asyncio.wait(workers)
    assert len(done) == 5  # sanity check
    data, stat = await zk.get(path)
    await zk.delete(path)
    assert int(data) == 5
    assert stat.version == 5


@pytest.mark.asyncio
async def test_counter_single_reused(zk, path):
    counter = zk.recipes.Counter(path)
    await counter.start()

    async def worker():
        await counter.incr()

    workers = []
    for _i in range(5):
        workers.append(worker())

    done, _pending = await asyncio.wait(workers)
    assert len(done) == 5
    val = await counter.get_value()
    await zk.delete(path)
    assert val == 5



import asyncio
import time

from .sequential import SequentialRecipe


class LeaderElection(SequentialRecipe):

    def __init__(self, base_path):
        super(LeaderElection, self).__init__(base_path)
        self.has_leadership = False

        self.leadership_future = self.client.loop.create_future()

    async def join(self):
        await self.create_unique_znode("candidate")
        await self.check_position()

    async def check_position(self, _=None):
        owned_positions, candidates = await self.analyze_siblings()
        if "candidate" not in owned_positions:
            return

        position = owned_positions["candidate"]

        self.has_leadership = bool(position == 0)

        if self.has_leadership:
            self.leadership_future.set_result(None)
            return

        await self.wait_on_sibling(candidates[position - 1])
        asyncio.ensure_future(self.check_position(), loop=self.client.loop)

    async def wait_for_leadership(self, timeout=None):
        if self.has_leadership:
            return

        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        if time_limit:
            await asyncio.wait_for(self.leadership_future, time_limit, loop=self.client.loop)
        else:
            await self.leadership_future

    async def resign(self):
        await self.delete_unique_znode("candidate")



import asyncio
import uuid

import pytest

from .. import WatchEvent
from ..exc import NoNode


@pytest.fixture
async def data_watcher(zk, path):
    await zk.create(path)
    watcher = zk.recipes.DataWatcher()
    watcher.set_client(zk)
    yield watcher
    await zk.delete(path)


@pytest.mark.asyncio
async def test_data_watch(zk, path, data_watcher):
    data = []
    ready = asyncio.Event()
    test_data = b'test' * 1000

    async def data_callback(d):
        data.append(d)
        ready.set()

    data_watcher.add_callback(path, data_callback)
    assert data == []
    await zk.set_data(path, test_data)
    await asyncio.wait_for(ready.wait(), timeout=0.1)
    assert ready.is_set()
    assert data == [test_data]
    data_watcher.remove_callback(path, data_callback)


@pytest.mark.asyncio
async def test_data_watch_delete(zk, path, data_watcher):
    data = []
    ready = asyncio.Event()
    test_data = b'test'

    async def data_callback(d):
        data.append(d)
        ready.set()

    await zk.set_data(path, test_data)

    data_watcher.add_callback(path, data_callback)
    await asyncio.sleep(0.2)
    assert data == [test_data]
    ready.clear()
    await zk.delete(path)

    await asyncio.wait_for(ready.wait(), timeout=1)
    assert ready.is_set()
    assert data == [test_data, NoNode]
    data_watcher.remove_callback(path, data_callback)

    await zk.create(path)


@pytest.mark.asyncio
async def test_data_watch_no_node(zk, path, data_watcher):
    random_path = path + uuid.uuid4().hex
    is_finished = asyncio.Future()

    async def stub_callback(d):
        assert d == NoNode
        is_finished.set_result(True)

    data_watcher.add_callback(random_path, stub_callback)
    await asyncio.wait_for(is_finished, 0.1)


@pytest.fixture
def child1(path):
    yield f'{path}/{uuid.uuid4().hex}'


@pytest.fixture
def child2(path):
    yield f'{path}/{uuid.uuid4().hex}'


@pytest.fixture
async def child_watcher(zk, path, child1, child2):
    await zk.create(path)
    watcher = zk.recipes.ChildrenWatcher()
    watcher.set_client(zk)
    yield watcher

    try:
        await zk.delete(child1)
        await zk.delete(child2)
    except NoNode:
        pass
    await zk.delete(path)


@pytest.mark.asyncio
async def test_child_watch(child_watcher, path, zk, child1, child2):
    children = set()
    ready = asyncio.Event()

    async def children_callback(c):
        for child in c:
            children.add(child)
            ready.set()

    child_watcher.add_callback(path, children_callback)
    assert children == set()
    await zk.create(child1)
    await asyncio.wait([ready.wait()], timeout=0.1)
    assert children == {child1.split('/')[-1]}
    ready.clear()
    await zk.create(child2)
    await asyncio.wait([ready.wait()], timeout=0.1)
    assert ready.is_set()
    assert children == {child.split('/')[-1] for child in (child1, child2)}
    child_watcher.remove_callback(path, children_callback)


@pytest.mark.asyncio
async def test_child_watch_no_node(child_watcher, path):
    random_path = path + uuid.uuid4().hex
    is_finished = asyncio.Future()

    async def stub_callback(d):
        assert d == NoNode
        is_finished.set_result(True)

    child_watcher.add_callback(random_path, stub_callback)
    await asyncio.wait_for(is_finished, 0.1)


@pytest.mark.asyncio
async def test_reconnect_watcher(data_watcher, path, zk_disruptor, zk, zk2):
    test_data = uuid.uuid4().hex.encode()
    ready = asyncio.Future()

    async def data_callback(d):
        print(f'Data callback get: {d}')
        if d == NoNode:
            return
        if d and not ready.done():
            print(f'Set result: {d} {ready}')
            ready.set_result(d)

    data_watcher.add_callback(path, data_callback)
    await zk_disruptor()
    await zk2.set_data(path, test_data)
    resp = await zk2.get_data(path)
    assert resp == test_data

    data = await asyncio.wait_for(ready, 1)
    assert data == test_data

    data_watcher.remove_callback(path, data_callback)


@pytest.mark.asyncio
async def test_watcher_fires_after_nonode(zk, data_watcher, child1):
    """
    Test that waiting for a nonexistent node is allowed if
    CREATED is in the watched events
    """
    messages = asyncio.Queue()
    data_watcher.watched_events.append(WatchEvent.CREATED)

    async def callback(d):
        print('Callback sees', d)
        await messages.put(d)

    # should trigger fetch right away, getting NoNode
    data_watcher.add_callback(child1, callback)

    no_node = await asyncio.wait_for(messages.get(), 1)
    assert no_node == NoNode

    # should trigger watch, which triggers fetch, which gets 'some data'
    await zk.create(child1, 'some data')
    some_data = await asyncio.wait_for(messages.get(), 1)
    assert some_data == b'some data'

    data_watcher.remove_callback(child1, callback)
    await zk.delete(child1)

@pytest.mark.asyncio
async def test_watcher_without_parents(zk, path, child1):
    """
    Make sure behavior is sane if ancestor node does not exist
    """
    final = f"{child1}/{uuid.uuid4().hex}"
    watcher = zk.recipes.DataWatcher(wait_for_create=True)
    messages = asyncio.Queue()

    async def callback(d):
        print('callback sees', d)
        await messages.put(d)

    watcher.add_callback(final, callback)

    # full path doesn't exist
    no_node = await asyncio.wait_for(messages.get(), 1)
    assert no_node == NoNode

    # create parent, no message should arrive
    await zk.create(path)
    await zk.create(child1)
    assert messages.empty() == True

    # create final node, should get 'howdy'
    await zk.create(final, b'howdy')
    howdy = await asyncio.wait_for(messages.get(), 1)
    assert howdy == b'howdy'

    watcher.remove_callback(final, callback)
    await zk.delete(final)
    await zk.delete(child1)
    await zk.delete(path)



import asyncio
import uuid

import pytest


class Tree:
    def __init__(self, zk, path):
        self.zk = zk
        self.path = path

    async def init(self):
        self.data = {}
        await self.zk.create(self.path)
        for name in ['node1', 'node2']:
            await self.zk.create(f'{self.path}/{name}')
            self.data[name] = {}
            for subname in ['subnode1', 'subnode2']:
                s = uuid.uuid4().hex.encode()
                await self.zk.create(f'{self.path}/{name}/{subname}', s)
                self.data[name][subname] = s

    async def modify(self):
        s = uuid.uuid4().hex
        new_node = f'{self.path}/{s}'
        await self.zk.create(new_node, s)
        self.data[s] = s.encode()

        sn1 = f'{self.path}/node1/subnode1'
        self.data['node1']['subnode1'] = s.encode()
        await self.zk.set_data(sn1, s)

        await self.zk.delete(f'{self.path}/node1/subnode2')
        del self.data['node1']['subnode2']
        await asyncio.sleep(0.1)


@pytest.fixture
async def tree(zk, path):
    out = Tree(zk, path)
    await out.init()
    yield out
    await zk.deleteall(path)


@pytest.fixture
async def tree_cache(zk, path):
    cache = zk.recipes.TreeCache(path)
    cache.set_client(zk)
    await cache.start()
    yield cache
    await cache.stop()


@pytest.mark.asyncio
async def test_tree_cache(zk, tree, tree_cache, path):
    assert tree_cache.as_dict() == tree.data

    td = b'test_data'
    await zk.set_data(f'{path}/node1', td)
    await asyncio.sleep(0.1)
    assert getattr(tree_cache.root, 'node1').value == td

    await tree.modify()
    assert tree_cache.as_dict() == tree.data


@pytest.mark.asyncio
async def test_stop_cachecrash(zk, path):
    cache = zk.recipes.TreeCache(path)
    cache.set_client(zk)
    await cache.start()
    await cache.stop()
    await zk.delete(path)


@pytest.mark.asyncio
async def test_cachecrash_after_delete(zk, path):
    cache = zk.recipes.TreeCache(path)
    cache.set_client(zk)
    await cache.start()
    await zk.delete(path)
    await cache.stop()



from .request import Request
from .response import Response


CLOSE_XID = None


class CloseRequest(Request):
    """
    """
    opcode = -11
    special_xid = CLOSE_XID

    parts = ()


class CloseResponse(Response):
    """
    """
    opcode = -11

    parts = ()



import logging
import asyncio

from aiozk import exc

from .data_watcher import DataWatcher
from .recipe import Recipe


log = logging.getLogger(__name__)


class Counter(Recipe):

    def __init__(self, base_path, default=0):
        super().__init__(base_path)

        self.value = None
        self.numeric_type = type(default)
        self._default = default
        self._version = 0

    async def _fetch(self):
        data, stat = await self.client.get(self.base_path)
        self._version = stat.version
        self.value = self.numeric_type(data)
        return (self.value, stat.version)

    async def start(self):
        base, _leaf = self.base_path.rsplit("/", 1)
        if base:
            await self.client.ensure_path(base)
        try:
            await self.client.create(self.base_path, str(self._default))
            self.value = self._default
            self._version = 0
        except exc.NodeExists:
            await self._fetch()

    async def get_value(self):
        data, _version = await self._fetch()
        return data

    async def set_value(self, value, force=True):
        data = str(value)
        version = -1 if force else self._version
        stat = await self.client.set(self.base_path, data, version)
        self.value = value
        self._version = stat.version
        log.debug("Set value to '%s': successful", data)

    async def apply_operation(self, operation):
        success = False
        while not success:
            new_value = operation(self.value)
            data = str(new_value)
            try:
                stat = await self.client.set(self.base_path, data, self._version)
                log.debug("Operation '%s': successful", operation.__name__)
                self.value = new_value
                self._version = stat.version
                success = True
            except exc.BadVersion:
                await self._fetch()
                log.debug(
                    "Operation '%s': version mismatch, retrying",
                    operation.__name__
                )

    async def incr(self, by=1):

        def increment(value):
            return value + self.numeric_type(by)

        await self.apply_operation(increment)

    async def decr(self, by=1):

        def decrement(value):
            return value - self.numeric_type(by)

        await self.apply_operation(decrement)

    def stop(self):
        """Deprecated. Counter no longer needs to be stopped"""
        pass



import asyncio
import collections
import logging
import random
import re

from aiozk import protocol, exc
from .connection import Connection
from .retry import RetryPolicy
from .states import States, SessionStateMachine

# from tornado import gen, ioloop


DEFAULT_ZOOKEEPER_PORT = 2181

MAX_FIND_WAIT = 60  # in seconds

HEARTBEAT_FREQUENCY = 3  # heartbeats per timeout interval


log = logging.getLogger(__name__)


class Session(object):

    def __init__(self, servers, timeout, retry_policy, allow_read_only, read_timeout, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.hosts = []
        for server in servers.split(","):
            ipv6_match = re.match(r'\[(.*)\]:(\d+)$', server)
            if ipv6_match is not None:
                host, port = ipv6_match.groups()
            elif ":" in server:
                host, port = server.rsplit(":", 1)
            else:
                host = server
                port = DEFAULT_ZOOKEEPER_PORT

            self.hosts.append((host, port))

        self.conn = None
        self.state = SessionStateMachine()

        self.retry_policy = retry_policy or RetryPolicy.forever()
        self.allow_read_only = allow_read_only

        self.xid = 0
        self.last_zxid = None

        self.session_id = None
        self.timeout = timeout
        self.password = b'\x00'
        self.read_timeout = read_timeout

        self.repair_loop_task = None

        self.heartbeat_handle = None

        self.watch_callbacks = collections.defaultdict(set)

        self.started = False
        self.closing = False

    async def ensure_safe_state(self, writing=False):
        safe_states = [States.CONNECTED]
        if self.allow_read_only and not writing:
            safe_states.append(States.READ_ONLY)

        if self.state in safe_states:
            return

        await self.state.wait_for(*safe_states, loop=self.loop)

    async def start(self):
        if self.started:
            await self.ensure_safe_state()
            return
        log.debug('Start session...')
        self.loop.call_soon(self.set_heartbeat)
        self.repair_loop_task = self.loop.create_task(self.repair_loop())
        self.repair_loop_task.add_done_callback(self._on_repair_loop_done)
        self.started = True
        await self.ensure_safe_state()

    def _on_repair_loop_done(self, repair_loop_task):
        if self.closing and repair_loop_task.cancelled():
            return

        if repair_loop_task.cancelled():
            log.error('Repair loop task cancelled when session is not in "closing" state')
            return

        repair_loop_exception = repair_loop_task.exception()
        if repair_loop_exception:
            log.error('Repair loop task failed with exception: {error}'.format(error=repair_loop_exception))

    async def find_server(self, allow_read_only):
        conn = None

        retry_policy = RetryPolicy.exponential_backoff(maximum=MAX_FIND_WAIT)

        while not conn:
            await retry_policy.enforce(loop=self.loop)

            servers = random.sample(self.hosts, len(self.hosts))
            for host, port in servers:
                log.info("Connecting to %s:%s", host, port)
                conn = await self.make_connection(host, port)
                if not conn or (conn.start_read_only and not allow_read_only):
                    continue
                log.info("Connected to %s:%s", host, port)
                break

            if not conn:
                log.warning("No servers available, will keep trying.")

        old_conn = self.conn
        self.conn = conn

        if old_conn:
            log.debug('Close old connection')
            self.loop.create_task(old_conn.close(self.timeout))

        if conn.start_read_only:
            self.loop.create_task(self.find_server(allow_read_only=False))

    async def make_connection(self, host, port):
        conn = Connection(host, port, watch_handler=self.event_dispatch, read_timeout=self.read_timeout, loop=self.loop)
        try:
            await conn.connect()
        except Exception:
            log.exception("Couldn't connect to %s:%s", host, port)
            return
        return conn

    async def establish_session(self):
        log.info("Establishing session. {!r}".format(self.session_id))
        connection_response = await self.conn.send_connect(
            protocol.ConnectRequest(
                protocol_version=0,
                last_seen_zxid=self.last_zxid or 0,
                timeout=int((self.timeout or 0) * 1000),
                session_id=self.session_id or 0,
                password=self.password,
                read_only=self.allow_read_only,
            )
        )
        if connection_response is None:
            # handle issue with inconsistent zxid on reconnection
            if self.state.current_state != States.LOST:
                self.state.transition_to(States.LOST)
            self.last_zxid = None
            raise exc.SessionLost()

        zxid, response = connection_response
        self.last_zxid = zxid

        if response.session_id == 0:  # invalid session, probably expired
            log.debug('Session lost')
            if self.state.current_state != States.LOST:
                self.state.transition_to(States.LOST)
            raise exc.SessionLost()

        log.info("Got session id %s", hex(response.session_id))
        log.info("Negotiated timeout: %s seconds", response.timeout / 1000)

        self.session_id = response.session_id
        self.password = response.password
        self.timeout = response.timeout / 1000

        self.last_zxid = zxid

    async def repair_loop(self):
        while not self.closing:
            log.debug('Await for repairable state')
            await self.state.wait_for(States.SUSPENDED, States.LOST, loop=self.loop)
            log.debug('Repair state reached')
            if self.closing:
                break

            await self.find_server(allow_read_only=self.allow_read_only)

            try:
                await asyncio.wait_for(self.establish_session(), self.timeout, loop=self.loop)
            except (exc.SessionLost, asyncio.TimeoutError) as e:
                log.info('Session closed: {}'.format(e))
                self.conn.abort(exc.SessionLost)
                await self.conn.close(self.timeout)  # TODO: make real timeout
                self.session_id = None
                self.password = b'\x00'
                continue

            if self.conn.start_read_only:
                self.state.transition_to(States.READ_ONLY)
            else:
                self.state.transition_to(States.CONNECTED)

            self.conn.start_read_loop()
            await self.set_existing_watches()

    async def send(self, request):
        response = None
        while not response:
            await self.retry_policy.enforce(request, loop=self.loop)
            await self.ensure_safe_state(writing=request.writes_data)

            try:
                self.xid += 1
                zxid, response = await self.conn.send(request, xid=self.xid)
                self.last_zxid = zxid
                self.set_heartbeat()
                self.retry_policy.clear(request)
            except (exc.NodeExists, exc.NoNode, exc.NotEmpty, exc.BadVersion):
                self.retry_policy.clear(request)
                raise
            except asyncio.CancelledError:
                self.retry_policy.clear(request)
                raise
            except exc.ConnectError:
                self.state.transition_to(States.SUSPENDED)
            except Exception as e:
                log.exception('Send exception: {}'.format(e))
                self.retry_policy.clear(request)
                raise e
        return response

    def set_heartbeat(self):
        timeout = self.timeout / HEARTBEAT_FREQUENCY
        if self.heartbeat_handle:
            self.heartbeat_handle.cancel()
        self.heartbeat_handle = self.loop.call_later(timeout, self.create_heartbeat)

    def create_heartbeat(self):
        self.loop.create_task(self.heartbeat())

    async def heartbeat(self):
        if self.closing:
            return

        await self.ensure_safe_state()

        try:
            timeout = self.timeout - self.timeout/HEARTBEAT_FREQUENCY
            zxid, _ = await asyncio.wait_for(self.conn.send(protocol.PingRequest()), timeout, loop=self.loop)
            self.last_zxid = zxid
        except (exc.ConnectError, asyncio.TimeoutError):
            self.state.transition_to(States.SUSPENDED)
        except Exception as e:
            log.exception('in heartbeat: {}'.format(e))
            raise e
        finally:
            self.set_heartbeat()

    def add_watch_callback(self, event_type, path, callback):
        self.watch_callbacks[(event_type, path)].add(callback)

    def remove_watch_callback(self, event_type, path, callback):
        self.watch_callbacks[(event_type, path)].discard(callback)
        if not self.watch_callbacks[(event_type, path)]:
            self.watch_callbacks.pop((event_type, path))

    def event_dispatch(self, event):
        log.debug("Got watch event: %s", event)

        if event.type:
            key = (event.type, event.path)
            for callback in self.watch_callbacks[key]:
                self.loop.call_soon(callback, event.path)
                # ioloop.IOLoop.current().add_callback(callback, event.path)
            return

        if event.state == protocol.WatchEvent.DISCONNECTED:
            log.error("Got 'disconnected' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.SESSION_EXPIRED:
            log.error("Got 'session expired' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.AUTH_FAILED:
            log.error("Got 'auth failed' watch event.")
            self.state.transition_to(States.LOST)
        elif event.state == protocol.WatchEvent.CONNECTED_READ_ONLY:
            log.warning("Got 'connected read only' watch event.")
            self.state.transition_to(States.READ_ONLY)
        elif event.state == protocol.WatchEvent.SASL_AUTHENTICATED:
            log.info("Authentication successful.")
        elif event.state == protocol.WatchEvent.CONNECTED:
            log.info("Got 'connected' watch event.")
            self.state.transition_to(States.CONNECTED)

    async def set_existing_watches(self):
        if not self.watch_callbacks:
            return

        request = protocol.SetWatchesRequest(
            relative_zxid=self.last_zxid or 0,
            data_watches=[],
            exist_watches=[],
            child_watches=[],
        )

        for event_type, path in self.watch_callbacks.keys():
            if event_type == protocol.WatchEvent.CREATED:
                request.exist_watches.append(path)
            if event_type == protocol.WatchEvent.DATA_CHANGED:
                request.data_watches.append(path)
            elif event_type == protocol.WatchEvent.CHILDREN_CHANGED:
                request.child_watches.append(path)

        await self.send(request)

    async def close(self):
        if not self.started:
            log.debug('Do nothing because session is not started')
            return
        if self.closing:
            return
        self.closing = True
        if self.repair_loop_task:
            self.repair_loop_task.cancel()
            await asyncio.wait_for(self.send(protocol.CloseRequest()), self.timeout,
                                   loop=self.loop)
        if self.state.current_state != States.LOST:
            self.state.transition_to(States.LOST)
        if self.conn:
            await self.conn.close(self.timeout)
        self.closing = False
        self.started = False



import asyncio
import time

from aiozk import exc, WatchEvent

from .recipe import Recipe


class Barrier(Recipe):

    def __init__(self, path):
        super(Barrier, self).__init__()
        self.path = path

    async def create(self):
        await self.ensure_path()
        await self.create_znode(self.path)

    async def lift(self):
        try:
            await self.client.delete(self.path)
        except exc.NoNode:
            pass

    async def wait(self, timeout=None):
        time_limit = None
        if timeout is not None:
            time_limit = time.time() + timeout

        barrier_lifted = self.client.wait_for_events(
            [WatchEvent.DELETED], self.path
        )

        exists = await self.client.exists(path=self.path, watch=True)
        if not exists:
            return

        try:
            if time_limit:
                await asyncio.wait_for(barrier_lifted, time_limit, loop=self.client.loop)
            else:
                await barrier_lifted
        except asyncio.TimeoutError:
            raise exc.TimeoutError



import asyncio
import pytest
from aiozk.exc import TimeoutError


@pytest.mark.asyncio
async def test_shared_lock(zk, path):

    shared = zk.recipes.SharedLock(path)
    got_lock = False
    async with await shared.acquire_write():
        got_lock = True

    assert got_lock

    got_released = False
    async with await shared.acquire_write():
        got_released = True

    assert got_released
    await zk.delete(path)


@pytest.mark.asyncio
async def test_shared_lock_timeout(zk, path):
    got_lock = False
    async with await zk.recipes.SharedLock(path).acquire_write():
        got_lock = True

        with pytest.raises(TimeoutError):
            await zk.recipes.SharedLock(path).acquire_write(timeout=1)

    assert got_lock

    got_released = False
    async with await zk.recipes.SharedLock(path).acquire_write():
        got_released = True

    assert got_released

    await zk.deleteall(path)



import asyncio

import asynctest
import pytest

import aiozk.connection


@pytest.fixture
def connection(event_loop):
    connection = aiozk.connection.Connection(
        host='zookeeper.test',
        port=2181,
        watch_handler=asynctest.MagicMock(),
        read_timeout=30,
        loop=asynctest.MagicMock(wraps=event_loop))

    connection.writer = asynctest.MagicMock()
    return connection


@pytest.mark.asyncio
async def test_close_connection_in_state_closing_do_not_performs_abort(connection):
    connection.abort = asynctest.CoroutineMock()
    connection.closing = True

    await connection.close(asynctest.ANY)

    connection.abort.assert_not_awaited()


@pytest.mark.asyncio
async def test_close_cancels_read_loop_task(connection):
    connection.start_read_loop()
    connection.read_response = asynctest.CoroutineMock(return_value=(0, asynctest.ANY, asynctest.ANY))

    task_cancelled_future = asyncio.Future()

    def set_result(task):
        task_cancelled_future.set_result(task.cancelled())

    connection.read_loop_task.add_done_callback(set_result)

    await connection.close(asynctest.ANY)
    assert await task_cancelled_future



import asyncio
from unittest import mock

import aiozk.session

import pytest
import asynctest

from aiozk import exc


@pytest.fixture
def session(event_loop):
    fake_retry_policy = asynctest.MagicMock(wraps=aiozk.session.RetryPolicy.forever())
    session = aiozk.session.Session(
        'zookeeper.test',
        timeout=10,
        retry_policy=fake_retry_policy,
        allow_read_only=True,
        read_timeout=30,
        loop=asynctest.MagicMock(wraps=event_loop),
    )
    session.state.transition_to(aiozk.session.States.CONNECTED)
    session.conn = asynctest.MagicMock()
    session.conn.send = asynctest.CoroutineMock()
    session.conn.close = asynctest.CoroutineMock()
    session.ensure_safe_state = asynctest.CoroutineMock()
    session.set_heartbeat = mock.Mock()
    return session


def prepare_repair_loop_callback_done_future(session):
    """Creates future that Awaits for session.repair_loop_task to be completed and returns value of session.closing"""
    future = asyncio.Future()
    repair_loop_task = session.repair_loop_task

    def set_result(_):
        future.set_result(session.closing)

    repair_loop_task.add_done_callback(set_result)
    return future


@pytest.mark.asyncio
async def test_start_session_twice(session):
    await session.start()
    session.ensure_safe_state.assert_called_once_with()
    session.ensure_safe_state.reset_mock()
    await session.start()
    session.ensure_safe_state.assert_called_once_with()

    session.loop.call_soon.assert_called_once()
    session.loop.create_task.assert_called_once()


@pytest.mark.asyncio
async def test_close_not_started(session):
    await session.close()
    assert not session.closing


@pytest.mark.asyncio
async def test_close_produces_no_error_log(session):
    session.conn.send.return_value = (mock.MagicMock(), mock.MagicMock())
    await session.start()

    repair_loop_callback_done = prepare_repair_loop_callback_done_future(session)
    with mock.patch.object(aiozk.session.log, 'error') as err_log_mock:

        assert not session.closing
        await session.close()

        session_closing = await repair_loop_callback_done

        assert session_closing
        assert not session.closing
        err_log_mock.assert_not_called()


@pytest.mark.asyncio
async def test_repair_loop_task_cancellation_produces_error_log(session):
    await session.start()

    repair_loop_callback_done = prepare_repair_loop_callback_done_future(session)
    with mock.patch.object(aiozk.session.log, 'error') as err_log_mock:
        session.repair_loop_task.cancel()

        with pytest.raises(asyncio.CancelledError):
            await session.repair_loop_task

        session_closing = await repair_loop_callback_done
        assert not session_closing
        err_log_mock.assert_called_once_with('Repair loop task cancelled when session is not in "closing" state')


@pytest.mark.asyncio
async def test_send_no_node(session):
    req = mock.MagicMock()
    session.conn.send.side_effect = exc.NoNode
    with pytest.raises(exc.NoNode):
        await session.send(req)

    session.retry_policy.clear.assert_called_once_with(req)
    session.conn.send.assert_called_once()
    session.set_heartbeat.assert_not_called()


@pytest.mark.asyncio
async def test_send_bad_version(session):
    req = mock.MagicMock()
    session.conn.send.side_effect = exc.BadVersion
    with pytest.raises(exc.BadVersion):
        await session.send(req)
    session.conn.send.assert_called_once()


@pytest.mark.asyncio
async def test_send_canceled(session):
    req = mock.MagicMock()
    session.conn.send.side_effect = asyncio.CancelledError
    with pytest.raises(asyncio.CancelledError):
        await session.send(req)

    session.retry_policy.clear.assert_called_once_with(req)
    session.conn.send.assert_called_once()
    session.set_heartbeat.assert_not_called()


@pytest.mark.asyncio
async def test_send_connection_error(session):
    zxid = 1
    session.conn.send.side_effect = [exc.ConnectError('zookeeper.test', '2181'), (zxid, mock.MagicMock())]
    req = mock.MagicMock()
    await session.send(req)

    session.retry_policy.clear.assert_called_once_with(req)
    assert session.conn.send.call_count == 2
    session.set_heartbeat.assert_called_once_with()


@pytest.mark.asyncio
async def test_send_unknown_error(session):
    req = mock.MagicMock()
    session.conn.send.side_effect = RuntimeError
    with pytest.raises(RuntimeError):
        await session.send(req)

    session.retry_policy.clear.assert_called_once_with(req)
    session.conn.send.assert_called_once()
    session.set_heartbeat.assert_not_called()


@pytest.mark.asyncio
async def test_send_good_case(session):
    zxid = 1
    req = mock.MagicMock()
    resp = mock.MagicMock()
    session.conn.send.return_value = (zxid, resp)
    await session.send(req)

    session.retry_policy.clear.assert_called_once_with(req)
    session.conn.send.assert_called_once()
    session.set_heartbeat.assert_called_once_with()
